# Made by Marcos Boggia
import cv2
from gui_automation.spot import Spot
from gui_automation.detector import Detector
import numpy as np
import imutils
from typing import Tuple


SQDIFF = cv2.TM_SQDIFF_NORMED
CCOEFF = cv2.TM_CCOEFF_NORMED
CCORR = cv2.TM_CCORR_NORMED


def obtain_tm_results(img, tpl, method) -> Tuple[int, Spot]:
    """
    Function that calls matchTemplate from OpenCV, normalizes similarity values and creates the Spot instance.

    :param tpl: image/numpy matrix with pixels. The image to be found.
    :param img: image/numpy matrix with pixels. The image which will be searched.
    :param method: opencv template match method to use: SQDIFF, CCOEFF, CCORR. Default is SQDIFF
    :return: similarity, Spot
    similarity level: 0(no similarity) - 1(exact match)
    Spot instance which wraps all positions/coordinates of the found image.
    """
    result = cv2.matchTemplate(img, tpl, method)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
    if method in [cv2.TM_SQDIFF, cv2.TM_SQDIFF_NORMED]:
        similarity = abs(min_val - 1)
        position = min_loc
    else:
        similarity = max_val
        position = max_loc
    return similarity, Spot(position, tpl.shape)


class TMDetector(Detector):
    """
    Class that wraps the template match detection.
    __init__() method has two optional parameters explained in docs.
    detect() method applies matchTemplate from OpenCV and returns similarity and coordinates wrapped in 'Spot' instance.
    """

    def __init__(self, method=SQDIFF, thresh=False):
        """
        :param method: opencv template match method to use: SQDIFF, CCOEFF, CCORR. Default is SQDIFF
        :param thresh: apply threshold binary filter.
        """
        self.method = method
        self.thresh = thresh

    @staticmethod
    def _apply_binary_thresh(tpl, img):
        _, tpl = cv2.threshold(tpl, 127, 255, cv2.THRESH_BINARY)
        _, img = cv2.threshold(img, 127, 255, cv2.THRESH_BINARY)
        return tpl, img

    # OpenCV Template Matching

    def detect(self, tpl, img):
        """
        Applies the matchTemplate to detect 'tpl' image over 'image' image.
        :param tpl: image/numpy matrix with pixels. The image to be found.
        :param img: image/numpy matrix with pixels. The image which will be searched.
        :return: similarity, Spot
        similarity level: 0(no similarity) - 1(exact match)
        Spot instance which wraps all positions/coordinates of the found image.
        """
        tpl = cv2.cvtColor(tpl, cv2.COLOR_BGR2GRAY)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if self.thresh:
            tpl, img = self._apply_binary_thresh(tpl, img)
        return obtain_tm_results(img, tpl, self.method)


class MultiscaledTMDetector(TMDetector):
    """ Child class of TMDetector. Same behaviour but detection occurs by resizing 'img' multiple times."""

    def __init__(self, method=SQDIFF, thresh=False, reduce_sc=0.2, magnify_sc=2.0, cant_sc=40):
        """
        :param method: opencv template match method to use: SQDIFF, CCOEFF, CCORR. Default is SQDIFF
        :param thresh: apply threshold binary filter.
        :param reduce_sc: size to downsize the image. Default is 0.2 --> 20% of the image.
        :param magnify_sc: size to upsize the image. Default is 2.0 --> 2x of the image.
        :param cant_sc: how many times the image will be resized within above values.
        """
        super().__init__(method, thresh)
        self.reduce_sc = reduce_sc
        self.magnify_sc = magnify_sc
        self.cant_sc = cant_sc

    def detect(self, tpl, img):
        """
        Applies the matchTemplate to detect 'tpl' image over 'image' image. 'image' gets resized in multiple scales.
        :param tpl: image/numpy matrix with pixels. The image to be found.
        :param img: image/numpy matrix with pixels. The image which will be searched.
        :return:
        """
        tpl = cv2.cvtColor(tpl, cv2.COLOR_BGR2GRAY)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if self.thresh:
            tpl, img = self._apply_binary_thresh(tpl, img)
        similarity, spot = obtain_tm_results(img, tpl, self.method)

        tpl_height, tpl_width = tpl.shape[:2]
        # Iterate through scales, resizing the image/screen and find better match for tpl.
        # Reducing up to reduce_sc. Default is 0.2, meaning it will resize image to a 0.2 of its size
        # Magnifying up to magnify_sc. Default is 2.0, meaning it will enlarge the image up to a double of its size
        for scale in np.linspace(self.reduce_sc, self.magnify_sc, self.cant_sc)[::-1]:
            # Resize
            resized = imutils.resize(img, width=int(img.shape[1] * scale))
            # if the resized image is smaller than the template, then break from the loop
            if resized.shape[0] < tpl_height or resized.shape[1] < tpl_width:
                break
            # apply tm
            new_similarity, new_spot = obtain_tm_results(resized, tpl, self.method)
            # keep the best match
            if new_similarity > similarity:
                similarity = new_similarity
                # Fix resized image coordinates.
                new_spot.revert_scaled_position_error(scale)
                spot = new_spot

        return similarity, spot



# Made by Marcos Boggia
from abc import ABC, abstractmethod


class Handler(ABC):
    @abstractmethod
    def screenshot(self, *args):
        pass

    def move(self, *args):
        pass

    def click(self, *args):
        pass

    def hold_click(self, *args):
        pass

    def drag_click(self, *args):
        pass



# Made by Marcos Boggia
from gui_automation.gui_automation import GuiAuto
from gui_automation.image_detector import TMDetector, MultiscaledTMDetector, SQDIFF, CCOEFF, CCORR
from gui_automation.foreground_handler import ForegroundHandler
from gui_automation.background_handler import BackgroundHandlerWin32
from gui_automation.image_loader import load_images
name = "gui_automation"



# Made by Marcos Boggia
import io
import cv2
import numpy as np
from PIL import Image
import win32gui
import win32ui
from ctypes import windll
from time import sleep
from gui_automation.handler import Handler


class BackgroundHandlerWin32(Handler):
    """
    Handler only for Windows. It can work on background(not viewable) apps. It needs the app name/window name.
    If needed you can pass the full hierarchy names as arguments to select a specific UI element.
    Eg: BackgroundHandlerWin32('window_name', 'main_ui_element', 'child_ui_element')
    If the element is not found raises ElementNotFound exception.
    """

    def __init__(self, app_name, *args):
        """

        :param app_name: app or window name of the desired app to be controlled.
        :param args: if you need to control a specific UI element, you can set these *args to determine the UI hierarchy
        names. Currently the condition asks if given name is 'in' the element name. Should be replaced with regexp in
        the future.
        """
        self.x_dis = 0
        self.y_dis = 0
        result = self.get_gui_elem(app_name, *args)
        if result:
            self.hwnd = result
            self.img_hwnd = result
        else:
            raise ElementNotFound

    def set_img_hwnd(self, app_name, *args, x_dis=0, y_dis=0):
        """
        It is possible that the handler used to control the app returns a black screen or similar.
        This function allows to set a different handler to get the screenshot of that GUI.
        Since there might be some positional difference between the original handler and the one used to obtain the
        image, x_dis and y_dis parameters can be used to fix these displacement.

        :param app_name: app or window name of the desired app to be controlled.
        :param args: if you need to control a specific UI element, you can set these *args to determine the UI hierarchy
        names. Currently the condition asks if given name is 'in' the element name. Should be replaced with regexp in
        the future.
        :param x_dis: displacement in pixels in between the handlers for the X axis.
        :param y_dis: same as x_dis but for the Y axis.
        :return:
        """
        result = self.get_gui_elem(app_name, *args)
        if result:
            self.img_hwnd = result
            self.x_dis = x_dis
            self.y_dis = y_dis
        else:
            raise ElementNotFound

    def screenshot(self):
        """
        Screenshot for background Win32 apps.
        :return: screen as OpenCV image format.
        """

        # OBTAIN IMAGE OF THE WINDOW SCREEN
        left, top, right, bot = win32gui.GetWindowRect(self.img_hwnd.handle)
        w = right - left
        h = bot - top
        hwnd_dc = win32gui.GetWindowDC(self.img_hwnd.handle)
        mfc_dc = win32ui.CreateDCFromHandle(hwnd_dc)
        save_dc = mfc_dc.CreateCompatibleDC()
        save_bit_map = win32ui.CreateBitmap()
        save_bit_map.CreateCompatibleBitmap(mfc_dc, w, h)
        save_dc.SelectObject(save_bit_map)
        result = windll.user32.PrintWindow(self.img_hwnd.handle, save_dc.GetSafeHdc(), 1)
        if result == 0:
            return False
        bmpinfo = save_bit_map.GetInfo()
        bmpstr = save_bit_map.GetBitmapBits(True)
        img = Image.frombuffer(
            'RGB',
            (bmpinfo['bmWidth'], bmpinfo['bmHeight']),
            bmpstr, 'raw', 'BGRX', 0, 1)
        win32gui.DeleteObject(save_bit_map.GetHandle())
        save_dc.DeleteDC()
        mfc_dc.DeleteDC()
        win32gui.ReleaseDC(self.img_hwnd.handle, hwnd_dc)
        # CONVERT IT TO OPENCV FORMAT
        img = cv2.cvtColor(np.array(img), cv2.COLOR_RGB2BGR)
        return img

    def click(self, x, y, clicks=1):
        for _ in range(clicks):
            print(f"HANDLER clicked x: {x + self.x_dis} y: {y + self.y_dis}")
            self.hwnd.click(coords=(x + self.x_dis, y + self.y_dis))

    def move(self, x, y):
        self.hwnd.move_mouse(coords=(x + self.x_dis, y + self.y_dis))

    def hold_click(self, x, y, time):
        refresh_rate = 0.05
        self.hwnd.press_mouse(coords=(x + self.x_dis, y + self.y_dis))
        while time > 0:
            sleep(refresh_rate)
            self.hwnd.move_mouse(coords=(x + self.x_dis, y + self.y_dis))
            time -= refresh_rate

        self.hwnd.release_mouse(coords=(x + self.x_dis, y + self.y_dis))

    def drag_click(self, start_x, start_y, end_x, end_y):
        self.hwnd.drag_mouse(press_coords=(start_x + self.x_dis, start_y + self.y_dis),
                             release_coords=(end_x + self.x_dis, end_y + self.y_dis))

    def press_key(self, key):
        print("Not implemented yet")

    def press_hotkey(self, *keys):
        print("Not implemented yet")

    def write_string(self, key):
        print("Not implemented yet")

    @staticmethod
    def get_gui_elem(app_name, *args):
        """

        :param app_name: name of the app. For eg: for Paint is MSPaintApp.
        :param args: name of the descendant elements ordered in hierarchy level.
        :return: returns the HWNDWrapper from pywinauto of the Gui element if found. Otherwise returns False.

        Due to some incompatibility between PyAutoGui and pywinauto, if pywinauto is imported, pyautogui stops working.
        That's why it is imported only when necessary(if this function is called).
        After this import, PyAutoGui will not continue to work. Which means foreground handler won't work neither.
        """
        from pywinauto.controls import hwndwrapper

        data = [elem for elem in args]
        data.insert(0, app_name)
        windows_list = []

        def _append(hwnd, _):
            windows_list.append(hwndwrapper.HwndWrapper(hwnd))

        win32gui.EnumWindows(_append, [])

        for elem in windows_list:
            result = BackgroundHandlerWin32.process_element(elem, data, 0)
            if result:
                return result
        return False

    @staticmethod
    def process_element(elem, data, i):
        """

        :param elem: HWNDWrapper element to start recursive search
        :param data: names of the elements ordered hierarchically.
        :param i: index of the data array.
        :return: returns the HWNDWrapper from pywinauto of the Gui element if found. Otherwise returns False.
        """
        name1 = elem.friendlyclassname
        name2 = elem.element_info.name
        if (name1 is not None and data[i] in name1) or (name2 is not None and data[i] in name2):
            if i == len(data) - 1:
                return elem
            else:
                children = elem.children()
                BackgroundHandlerWin32.load_children(children)
                if children:
                    for next_child in children:
                        hwnd = BackgroundHandlerWin32.process_element(next_child, data, i + 1)
                        if hwnd:
                            return hwnd
        return False

    @staticmethod
    def load_children(children):
        """
        WORKAROUND for children not being loaded. Rarely, printing them fix this issue. Maybe its some lazy loading.
        This supress stdout when printing.
        :param children:
        :return:
        """
        from contextlib import redirect_stdout
        trap = io.StringIO()
        with redirect_stdout(trap):
            print(children)


class ElementNotFound(Exception):
    pass



# Made by Marcos Boggia
import pyautogui
from time import sleep
import numpy as np
from pyautogui import screenshot
from gui_automation.handler import Handler

pyautogui.PAUSE = 1
pyautogui.FAILSAFE = False


class ForegroundHandler(Handler):
    """ Default handler. It screenshots the viewable screen and performs clicks on it."""
    @staticmethod
    def screenshot():
        return np.array(screenshot())[:, :, ::-1].copy()

    @staticmethod
    def click(x, y, clicks=1):
        pyautogui.click(x=x, y=y, button='left', clicks=clicks, interval=0.5)

    @staticmethod
    def move(x, y):
        pyautogui.moveTo(x=x, y=y)

    @staticmethod
    def hold_click(x, y, time):
        pyautogui.PAUSE = 0
        refresh_rate = 0.05
        pyautogui.mouseDown(x=x, y=y, button='left')
        while time > 0:
            sleep(refresh_rate)
            pyautogui.moveTo(x=x, y=y)
            time -= refresh_rate

        pyautogui.mouseUp(x=x, y=y, button='left')
        pyautogui.PAUSE = 1

    @staticmethod
    def drag_click(start_x, start_y, end_x, end_y):
        pyautogui.moveTo(x=start_x, y=start_y)
        pyautogui.dragTo(x=end_x, y=end_y, button='left', duration=1)

    @staticmethod
    def press_key(key):
        pyautogui.press(key)

    @staticmethod
    def press_hotkey(*keys):
        pyautogui.hotkey(*keys)

    @staticmethod
    def write_string(key):
        pyautogui.write(key)

